'''
    File: lineup.py
    Author: Drew Scott
'''

import pkg_resources
from typing import List, Optional
import random

from .player import Player

class Lineup:
    '''
        Stores the players in a Lineup

        Public methods:
            set_players: sets players in lineup either randomly or based on input filename
            get_player: returns a Player instance who matches first_name and last_name from
                the players list
    '''

    def __init__(self):
        self.players: List[Player] = []
        self.random_lineup = True

    def set_players(self, lineup_filename: Optional[str] = None) -> None:
        '''
            Sets the players in this lineup
        '''

        # clear players
        self.players = []

        # add the players
        if lineup_filename is None:
            self._generate_random_lineup()
            self.random_lineup = False
        else:
            self._generate_file_lineup(lineup_filename)
            self.random_lineup = True

        # display the players
        print("Players:")
        for player in self.players:
            print(player)
        print()

    def get_player(self, first_name: str, last_name: str) -> Player:
        '''
            Returns the player in the lineup who matches first and last name
        '''

        for player in self.players:
            if player.player_info['first_name'] == first_name \
                and player.player_info['last_name'] == last_name:
                return player

        raise Exception(f'Player {first_name} {last_name} not found from PA outcome file')

    def _generate_random_lineup(self) -> None:
        '''
            Generates a random lineup of 9 players drawn from stats_filename
        '''

        stats = pkg_resources.resource_stream(__name__, Player.stats_filepath).read().decode(encoding='utf-8-sig')
        stat_lines = stats.split('\n')

        # get all the players
        players = []
        for line in stat_lines[1:]:
            players.append(Player(line))

        # select 9 random players to add to the lineup
        player_indexes = self._get_nine(len(players))
        for index in player_indexes:
            self._add_player(players[index])

    def _generate_file_lineup(self, lineup_filename: str) -> None:
        '''
            Sets the players in the lineup file to this lineup, in the order listed in the file
        '''

        # TODO: what if lineup_filename not located in package contents (i.e. is user generated?)
        lineup_filepath = Player.data_directory + Player.lineups_directory + lineup_filename

        # get the players specified in the input file
        raw_players = pkg_resources.resource_stream(__name__, lineup_filepath).read().decode().split('\n')[:-1]
        player_names = []
        for player in raw_players:
            first, last = player.split()
            player_names.append(f'{last},{first}')

        # read the player data from the stats csv
        stats = pkg_resources.resource_stream(__name__, Player.stats_filepath).read().decode(encoding='utf-8-sig')
        stat_lines = stats.split('\n')

        players: List[Optional[Player]] = [None] * 9 
        for line in stat_lines[1:]:
            splits = line.split(',')
            name = splits[0] + ',' + splits[1]

            if name in player_names:
                players[player_names.index(name)] = Player(line)

        for player in players:
            self._add_player(player)

        if len(self.players) != 9:
            raise Exception(f'Incorrect number of players: {len(lineup.players)}, {lineup.players}')

    def _add_player(self, player: Optional[Player]) -> None:
        '''
            Adds a player to the lineup
        '''

        if player is None:
            raise Exception('Tried adding None to lineup')

        self.players.append(player)

    @staticmethod
    def _get_nine(total_count: int) -> List[int]:
        '''
            Returns a list of length 9 with unique indexes in the range of total_count
        '''
        nine: List[int] = []

        while len(nine) < 9:
            index = random.randint(0, total_count - 1)
            if index not in nine:
                nine.append(index)

        return nine



'''
    File: simulation.py
    Author: Drew Scott
'''

import pkg_resources
from statistics import mean
import random
from typing import List, Tuple, Optional, Dict
import heapq
import itertools
from multiprocessing import Pool, cpu_count
from functools import partial
import tqdm # type: ignore

from .lineup import Lineup
from .player import Player

class Simulation:
    '''
        Public methods:
            run_full_sim: runs sims_per_order games through every possible order of the lineup
    '''

    def __init__(self, lineup: Lineup, sims_per_order: int = 1, pa_outcomes_filename: Optional[str] = None):
        self.sims_per_order = sims_per_order
        self.lineup = lineup

        self.pa_outcomes = []
        self._set_pa_outcomes(pa_outcomes_filename)   

    def run_full_sim(self) -> None:
        '''
            Simulates per_order games for each possible batting order
        '''

        orders = list(itertools.permutations(range(9)))

        # run the simulation
        avg_runs_per_order = []

        num_cores = cpu_count()
        with Pool(num_cores) as pool:
            avg_runs_per_order = list(tqdm.tqdm( \
                pool.imap(partial(self._sim_order), orders), \
                total=len(orders)))

        # aggregate data
        avg_runs_indexes = [(runs, i) for i, runs in enumerate(avg_runs_per_order)]
        best_five = heapq.nlargest(5, avg_runs_indexes)
        worst_five = heapq.nsmallest(5, avg_runs_indexes)

        best_runs = max(avg_runs_per_order)
        worst_runs = min(avg_runs_per_order)
        avg_runs = mean(avg_runs_per_order)

        # print results
        print(f'\nTotal games simulated: {len(orders) * self.sims_per_order:,}')
        print(f'Games simulated per order: {self.sims_per_order:,}')
        print(f'Total orders simulated: {len(orders):,}')
        print()
        print(f'Max runs for sim: {best_runs}')
        print(f'Avg. runs for sim: {avg_runs:.2f}')
        print(f'Min runs for sim: {worst_runs}\n')

        if not self.lineup.random_lineup:
            print('Order of interest:')
            print(f'Average runs for order: {avg_runs_per_order[0]}')
            for ind in orders[0]:
                print(f'\t{str(self.lineup.players[ind])}')

        print('\nTop 5 batting orders:')
        for rank in range(len(best_five)):
            order = orders[best_five[rank][1]]
            print(f'{str(rank+1)}) Average runs for order: {best_five[rank][0]}')
            for ind in order:
                print(f'\t{str(self.lineup.players[ind])}')
            print()

        print('Bottom 5 batting orders:')
        for rank in range(len(worst_five)-1, -1, -1):
            order = orders[worst_five[rank][1]]
            print(f'{str(len(orders) - rank)}) Average runs for order: {worst_five[rank][0]}')
            for ind in order:
                print(f'\t{str(self.lineup.players[ind])}')
            print()

    def _sim_order(self, order: List[int]) -> float:
        '''
            Simulates per_order games for the given order
        '''

        tot_runs_order = 0
        for game_num in range(self.sims_per_order):
            leadoff = 0
            thru_order = 0
            pa_outcomes_game = self.pa_outcomes[game_num]
            for _ in range(9):
                runs, leadoff, thru_order = self._sim_inning(\
                    pa_outcomes_game, leadoff, thru_order, order)

                tot_runs_order += runs

        avg_runs_order = tot_runs_order / self.sims_per_order

        return avg_runs_order

    def _sim_inning(
            self,
            generated_outcomes: Dict[int, List[int]],
            leadoff: int,
            thru_order: int,
            order: List[int],
        ) -> Tuple[int, int, int]:

        '''
            Simulates an inning of play
            Returns the number of runs scored in the inning and what batter will lead off next inning
        '''

        runs = 0
        outs = 0
        runners: List[Optional[Player]] = [None, None, None]

        cur_batter_pos = leadoff
        cur_batter = order[cur_batter_pos]
        cur_player = self.lineup.players[cur_batter]

        while outs < 3:
            # TODO: simulate stealing and other mid-PA events

            # there aren't enough outcomes for this batter, so guarantee an out
            # TODO: simulate PA -- should do this in Player class so it holds constant across sims
            if len(generated_outcomes[cur_batter]) <= thru_order:
                outs += 1
                cur_batter_pos, thru_order = self._new_batter(cur_batter_pos, thru_order)
                cur_batter = order[cur_batter_pos]
                continue

            outcome = generated_outcomes[cur_batter][thru_order]

            # TODO: make it so the bit shifting can handle changes to outcomes
            # and how much info is encoded in outcome int
            if (8 << 2 <= outcome < 12 << 2) or (4 << 2 <= outcome < 5 << 2):
                # out on fly, ground, ld, or popup or strikeout

                # need to account for sacrficies and double/triple plays
                outs += 1
                if outs == 3:
                    # end of inning, no sacs possible
                    break

                # TODO: actually find rates
                if outcome == 'b_out_fly':
                    # sacrifice rates
                    third_scores = 0.2
                    second_advances = 0.15
                    if outcome & 0b11 == 2:
                        # hit to right
                        second_advances = 0.3
                    elif outcome & 0b11 == 0:
                        # hit to left
                        second_advances = 0.05

                    sac_rand = random.random()
                    if sac_rand < third_scores and runners[2] is not None:
                        runs += 1
                        runners = [runners[0], runners[1], None]

                    if sac_rand < second_advances and runners[1] is not None and runners[2] is None:
                        runners = [runners[0], None, runners[1]]

                elif outcome == 'b_out_ground':
                    sac_rate = 0.001
                    double_play_rate = 0.3

                    if random.random() < sac_rate:
                        runs += int(runners[2] is not None)
                        runners = [None, runners[0], runners[1]]
                    elif random.random() < double_play_rate:
                        outs += int(runners[0] is not None)
                        runners = [None, runners[1], runners[2]]

                        if outs >= 3:
                            break

            elif 5 << 2 <= outcome < 8 << 2:
                # walk, catcher interference, or hbp
                # right now treating all these the same:
                # just cascade runner advancements along starting with the batter
                if runners[0] is not None:
                    if runners[1] is not None:
                        if runners[2] is not None:
                            # runners on first, second, and third
                            runs += 1
                            runners = [cur_player, runners[0], runners[1]]

                        else:
                            # runners on first and second
                            runners = [cur_player, runners[0], runners[1]]
                    else:
                        # runner on first, maybe third
                        runners = [cur_player, runners[0], runners[2]]
                else:
                    # no runner on first
                    runners = [cur_player, runners[1], runners[2]]

            # TODO: account for direction on hits for base runner advancements
            elif 0 << 2 <= outcome < 1 << 2:
                # on a single, runners on second and third score, and runner on first goes to second
                runs += int(runners[1] is not None) + int(runners[2] is not None)
                runners = [cur_player, runners[0], None]

            elif 1 << 2 <= outcome < 2 << 2:
                # on a double, a runner on first advances to third, batter to second, and others score
                runs += int(runners[1] is not None) + int(runners[2] is not None)
                runners = [None, cur_player, runners[0]]

            elif 2 << 2 <= outcome < 3 << 2:
                # all runners score, batter to third
                runs += int(runners[0] is not None) + int(runners[1] is not None) + \
                    int(runners[2] is not None)
                runners = [None, None, cur_player]

            elif 3 << 2 <= outcome < 4 << 2:
                # all runners and batter score
                runs += int(runners[0] is not None) + int(runners[1] is not None) + \
                    int(runners[2] is not None) + 1
                runners = [None, None, None]
            else:
                raise Exception(f"Invalid outcome name: {outcome}")

            cur_batter_pos, thru_order = self._new_batter(cur_batter_pos, thru_order)
            cur_batter = order[cur_batter_pos]
            cur_player = self.lineup.players[cur_batter]

        cur_batter_pos, thru_order = self._new_batter(cur_batter_pos, thru_order)

        return runs, cur_batter_pos, thru_order

    def _set_pa_outcomes(self, outcome_filename: Optional[str]) -> None:
        ''' 
            Generates the PA outcomes for each player in the lineup for each game that will be simulated
            If an outcome filename is supplied, outcomes will be retreived from there, and sims_per_order ignored
            Stores this information both in the Player instance and in this Simulation instance
        '''

        # set the outcomes in all of the player instances
        if outcome_filename is None:
            for player in self.lineup.players:
                player.generate_pa_outcomes(self.sims_per_order)

        else:
            # TODO: good place to do this?
            self.sims_per_order = 1

            outcome_filepath = Player.data_directory + Player.outcomes_directory + outcome_filename
            raw_outcomes = pkg_resources.resource_stream(__name__, outcome_filepath).read().decode().split('\n')[:-1]

            for player_outcome in raw_outcomes:
                first_name, last_name = player_outcome.split(':')[0].split()
                player = self.lineup.get_player(first_name, last_name)

                player.set_pa_outcomes(player_outcome.split(':')[1].split(','))

        # read the outcomes from the player instances into the simulation instance
        for game_num in range(self.sims_per_order):
            game_pas = {}
            for i, player in enumerate(self.lineup.players):
                game_pas[i] = player.pa_outcomes[game_num]

            self.pa_outcomes.append(game_pas)

    @staticmethod
    def _new_batter(cur_batter: int, thru_order: int) -> Tuple[int, int]:
        '''
            Returns the index of the next batter
        '''
        if cur_batter == 8:
            return 0, thru_order+1

        return cur_batter + 1, thru_order



'''
    File: Lineup.py
    Author: Drew Scott
'''

import pkg_resources
import random
from typing import List, Dict, Union, Any

class Player:
    '''
        Public methods:
            generate_pa_outcomes: generates n_games pa outcomes based on the player's stats
            set_pa_outcomes: set the player's pa outcomes to the input (only one game allowed)
    '''

    col_names: List[str] = []

    pa_outcome_names = ['b_single', 'b_double', 'b_triple', 'b_home_run', 'b_strikeout', 'b_walk', \
                        'b_catcher_interf', 'b_hit_by_pitch', 'b_out_fly', 'b_out_ground', \
                        'b_out_line_drive', 'b_out_popup']

    outcome_name_to_num_mapping: Dict[str, int] = {}

    pa_outcome_directions = ['pull_percent', 'straightaway_percent', 'opposite_percent']

    longest_name: int = 0

    metadata_isset = False

    # main data directory
    data_directory = 'data/' 

    # in main data directory
    lineups_directory = 'teams/'
    outcomes_directory = 'pa_outcomes/'
    stats_filename = 'stats.csv'

    # full paths
    stats_filepath = data_directory + stats_filename

    def __init__(self, stat_line: str):
        if not Player.metadata_isset:
            Player._set_metadata()

        self.player_info: Dict[str, Any] = {} 
        self._set_player_info(stat_line)

        self.player_pa_probabilities: Dict[str, float] = {} 
        self._set_player_pa_probabilities()        

        self.player_pa_thresholds: List[float] = [] 
        self._set_player_pa_thresholds()

        self.pa_outcomes: List[List[int]] = []

    def generate_pa_outcomes(self, n_games: int) -> None:
        '''
            Sets pa_outcomes to have 10 outcomes per game, generated by the player's thresholds
        '''

        generated_outcomes = []

        for _ in range(n_games):
            game_outcomes = []
            for _ in range(10):
                outcome = self._get_outcome_name()
                # generate a direction,
                # even if it's not possible (like a strikeout) since it will be ignored
                direction = self._get_outcome_direction()

                game_outcomes.append(self._full_outcome_to_num(outcome, direction))

            generated_outcomes.append(game_outcomes)

        self.pa_outcomes = generated_outcomes

    def set_pa_outcomes(self, pa_outcomes: List[str]) -> None:
        '''
            Sets pa_outcomes to the given outcomes, only one game allowed
            (used when simulating a single game with pre-determined outcomes)
        '''

        # TODO: acutally simulate direction
        self.pa_outcomes = [[self._full_outcome_to_num(outcome, 'C') for outcome in pa_outcomes]]

    def _get_outcome_name(self) -> str:
        '''
            Returns the name of the plate appearance outcome based on a randomly generated number
        '''

        rand_outcome = random.random()
        for out_threshold, outcome in zip(self.player_pa_thresholds, Player.pa_outcome_names):
            if rand_outcome < out_threshold:
                return outcome

        return Player.pa_outcome_names[-1]

    def _get_outcome_direction(self) -> str:
        '''
            Generates the direction of the PA outcome
        '''

        # determine which relative direction the outcome was
        # TODO: make better
        direction_num = random.random() * 100
        relative_direction = Player.pa_outcome_directions[2]
        if direction_num < float(self.player_info[Player.pa_outcome_directions[0]]):
            relative_direction = Player.pa_outcome_directions[0]
        elif direction_num < float(self.player_info[Player.pa_outcome_directions[0]]) + \
                float(self.player_info[Player.pa_outcome_directions[1]]):
            relative_direction = Player.pa_outcome_directions[1]

        # TODO: get pitcher info
        # assume pitcher is a righty, so anyone who switch hits will be lefty
        batter_side = 'R'
        if float(self.player_info['bats_left_pct']) > 0:
            batter_side = 'L'

        # turn relative direction into true direction
        true_direction = 'C'
        if relative_direction == 'pull_percent':
            if batter_side == 'R':
                true_direction = 'L'
            else:
                true_direction = 'R'
        elif relative_direction == 'opposite_percent':
            true_direction = batter_side

        return true_direction

    def _set_player_info(self, stat_line: str) -> None:
        '''
            Sets all of the player's information
        '''

        stat_splits = stat_line[: -1].split(',')

        stats = [int(s) if s.isnumeric() \
                    else float(s) if Player._isfloat(s) \
                    else s \
                for s in stat_splits]

        for col, stat in zip(Player.col_names, stats):
            self.player_info[col] = stat

    def _set_player_pa_probabilities(self) -> None:
        '''
            Sets the probability of each possible plate appearance outcomes for the input player
        '''

        total_pa = int(self.player_info['b_total_pa'])

        for outcome, outcome_count in self.player_info.items():
            if outcome not in Player.pa_outcome_names:
                continue

            prob = int(outcome_count) / total_pa
            self.player_pa_probabilities[outcome] = prob

    def _set_player_pa_thresholds(self) -> None:
        '''
            Sets the upper threshold (between 0 and 1) for each outcome name.
            Used to decide a plate appearance outcome using a random number
        '''

        self.player_pa_thresholds.append(self.player_pa_probabilities[Player.pa_outcome_names[0]])
        for i, outcome in enumerate(Player.pa_outcome_names[1:], start=1):
            self.player_pa_thresholds.append(self.player_pa_thresholds[i-1] + self.player_pa_probabilities[outcome])

    def _name_length(self) -> int:
        '''
            Returns the length of the player's name
        '''

        return len(self.player_info['first_name']) + len(self.player_info['last_name'])

    @staticmethod
    def _full_outcome_to_num(outcome_name: str, direction: str) -> int:
        '''
            Converts the outcome information to a number
        '''

        num = 0

        #outcome_bits = math.ceil(math.log(len(Player.pa_outcome_names), 2))
        outcome_num = Player.outcome_name_to_num_mapping[outcome_name]

        direction_num = 0
        if direction == 'C':
            direction_num = 1
        elif direction == 'R':
            direction_num = 2

        num = (outcome_num << 2) + direction_num

        return num

    @staticmethod
    def _isfloat(string: str) -> bool:
        '''
            Returns true if a str can be read as a float, false otherwise
        '''

        if string.isnumeric():
            return True

        splits = string.split('.')
        if len(splits) == 1 or len(splits) == 2:
            if len(splits[0]) == 0:
                return False

            if splits[0][0] == '-':
                splits[0] = splits[0][1:]

            if len(splits) == 2:
                return splits[0].isnumeric() and splits[1].isnumeric()

            return splits[0].isnumeric()

        return False

    @classmethod
    def _set_metadata(cls) -> None:
        '''
            Sets up some static metadata for the Player class
        '''

        stats = pkg_resources.resource_stream(__name__, Player.stats_filepath).read().decode(encoding='utf-8-sig')
        stat_lines = stats.split('\n')

        # get all of the column names
        cls.col_names = stat_lines[0].split(',')

        # get the maximum name length
        cls.longest_name = 0
        for line in stat_lines[1:]:
            last_length = line.index(',')
            first_length = line[last_length+1:].index(',')

            cls.longest_name = max(cls.longest_name, last_length + first_length + 1)

        # map each outcome name to a number
        for i, outcome in enumerate(cls.pa_outcome_names):
            cls.outcome_name_to_num_mapping[outcome] = i

        # mark metadata set
        Player.metadata_isset = True

    def __repr__(self):
        tab_size = 8
        num_tabs = ((Player.longest_name + 1 - self._name_length()) // tab_size) + 1
        return f'{self.player_info["first_name"]} {self.player_info["last_name"]}:' + \
                '\t'*num_tabs + \
                f'{self.player_info["on_base_plus_slg"]:.3f} OPS, '+ \
                f'{self.player_info["woba"]:.3f} WOBA'
